/* 
 * File:     PaeCamera.cpp
 * Author:   PAE
 * Date:     18.03.2006
 * Modified: 23.03.2006
 *
 * Purpose: Sourcecode zur Klasse PaeCamera
 */
 
#include "PaeCamera.h"
#include "Utilities.h"

//routinen zu der kamera klasse


PaeCamera::PaeCamera(void){
  pos[0] =  0.0;
  pos[1] =  0.0;
  pos[2] = 10.0;
  
  up[0]  =  0.0;
  up[1]  =  0.0;
  up[2]  =  1.0;
  
  dir[0] =  sqrt(0.5);
  dir[1] =  sqrt(0.5);
  dir[2] =  0.0;

  rig[0] = -sqrt(0.5);
  rig[1] =  sqrt(0.5);
  rig[2] =  0.0;
}

void PaeCamera::SetCam(void){
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  gluLookAt(pos[0], pos[1], pos[2], 
       pos[0]+dir[0], pos[1]+dir[1], pos[2]+dir[2], 
       up[0], up[1], up[2]);
}


void PaeCamera::RotateDir(float angle){
  GLfloat v[2], len;
  v[0] = dir[0];
  v[1] = dir[1];
  dir[0] = cos(angle) * v[0] + sin(angle) * v[1];
  dir[1] = cos(angle) * v[1] - sin(angle) * v[0];
  
  v[0] = up[0];
  v[1] = up[1];
  up[0] = cos(angle) * v[0] + sin(angle) * v[1];
  up[1] = cos(angle) * v[1] - sin(angle) * v[0];


  /*
   * recompute the right vector
   */

  rig[0] = up[1] * dir[2] - up[2] * dir[1];  
  rig[1] = up[2] * dir[0] - up[0] * dir[2];
  rig[2] = up[0] * dir[1] - up[1] * dir[0];
  
  len = sqrt(rig[0]*rig[0] + rig[1]*rig[1] + rig[2]*rig[2]);
  
  rig[0] /= len;
  rig[1] /= len;
  rig[2] /= len;
  

}


void PaeCamera::StatusPrint(void){

  cout << " = = = = = = = = = = = = = = = =" << endl;

  cout << "Position: " << endl;
  cout << "   pos[0] = " << pos[0] << endl;
  cout << "   pos[1] = " << pos[1] << endl;
  cout << "   pos[2] = " << pos[2] << endl;

  cout << "Direction: " << endl;
  cout << "   dir[0] = " << dir[0] << endl;
  cout << "   dir[1] = " << dir[1] << endl;
  cout << "   dir[2] = " << dir[2] << endl;

  cout << "Up-Vector: " << endl;
  cout << "    up[0] = " <<  up[0] << endl;
  cout << "    up[1] = " <<  up[1] << endl;
  cout << "    up[2] = " <<  up[2] << endl;

  cout << "Right-Vector: " << endl;
  cout << "   rig[0] = " << rig[0] << endl;
  cout << "   rig[1] = " << rig[1] << endl;
  cout << "   rig[2] = " << rig[2] << endl;

  cout << " = = = = = = = = = = = = = = = =" << endl;
  
}






void PaeCamera::Rotate(float length){

  float temp;

  dir[0] += (up[0] * length);
  dir[1] += (up[1] * length);
  dir[2] += (up[2] * length);
  
  temp = sqrt(dir[0]*dir[0] + dir[1]*dir[1] + dir[2]*dir[2]);
  
  dir[0] /= temp;
  dir[1] /= temp;
  dir[2] /= temp;
  
  up[0] = -rig[1] * dir[2] + rig[2] * dir[1];
  up[1] = -rig[2] * dir[0] + rig[0] * dir[2];
  up[2] = -rig[0] * dir[1] + rig[1] * dir[0];
  
  temp = sqrt(up[0]*up[0] + up[1]*up[1] + up[2]*up[2]);
  
  up[0] /= temp;
  up[1] /= temp;
  up[2] /= temp;

  rig[0] = up[1] * dir[2] - up[2] * dir[1];  
  rig[1] = up[2] * dir[0] - up[0] * dir[2];
  rig[2] = up[0] * dir[1] - up[1] * dir[0];
}



void PaeCamera::Pitch(float length){

  GLfloat len;

  /*
   * Compute the new direction
   */

  dir[0] += length*up[0];
  dir[1] += length*up[1];
  dir[2] += length*up[2];

  len = sqrt(dir[0]*dir[0]+dir[1]*dir[1]+dir[2]*dir[2]);

  dir[0] /= len;
  dir[1] /= len;
  dir[2] /= len;

  /*
   * recompute the new up vector
   */

  up[0] = dir[1] * rig[2] - dir[2] * rig[1];  
  up[1] = dir[2] * rig[0] - dir[0] * rig[2];
  up[2] = dir[0] * rig[1] - dir[1] * rig[0];
}


void PaeCamera::Yaw(float length){

  GLfloat len;

  /*
   * Compute the new direction
   */

  dir[0] += length*rig[0];
  dir[1] += length*rig[1];
  dir[2] += length*rig[2];

  len = sqrt(dir[0]*dir[0]+dir[1]*dir[1]+dir[2]*dir[2]);

  dir[0] /= len;
  dir[1] /= len;
  dir[2] /= len;

  /*
   * recompute the new right vector
   */

  rig[0] = up[1] * dir[2] - up[2] * dir[1];  
  rig[1] = up[2] * dir[0] - up[0] * dir[2];
  rig[2] = up[0] * dir[1] - up[1] * dir[0];
}


void PaeCamera::Roll(float length){

  GLfloat len;

  /*
   * Compute the new up
   */

  up[0] += length*rig[0];
  up[1] += length*rig[1];
  up[2] += length*rig[2];

  len = sqrt(up[0]*up[0]+up[1]*up[1]+up[2]*up[2]);

  up[0] /= len;
  up[1] /= len;
  up[2] /= len;

  /*
   * recompute the new right vector
   */

  rig[0] = up[1] * dir[2] - up[2] * dir[1];  
  rig[1] = up[2] * dir[0] - up[0] * dir[2];
  rig[2] = up[0] * dir[1] - up[1] * dir[0];
}





void PaeCamera::MoveUp(float length){
  pos[0] += (up[0] * length);
  pos[1] += (up[1] * length);
  pos[2] += (up[2] * length);
}





void PaeCamera::MoveForward(float length){
  pos[0] += (dir[0] * length);
  pos[1] += (dir[1] * length);
  pos[2] += (dir[2] * length);
}



void PaeCamera::MoveSide(float length){
  pos[0] += (rig[0] * length);
  pos[1] += (rig[1] * length);
  pos[2] += (rig[2] * length);
}



void PaeCamera::SetPosition(float x, float y, float z){

  pos[0] = x;
  pos[1] = y;
  pos[2] = z;

}

void PaeCamera::SetDirection(float dx, float dy, float dz){

  (void) PaeRenorm3(dir);

  dir[0] = dx;
  dir[1] = dy;
  dir[2] = dz;

  up[0] = -dz;
  up[1] = 0.0;
  up[2] =  dx;
  
  /*
   * recompute the new right vector
   */

  rig[0] = up[1] * dir[2] - up[2] * dir[1];  
  rig[1] = up[2] * dir[0] - up[0] * dir[2];
  rig[2] = up[0] * dir[1] - up[1] * dir[0];
}

