/* 
 * File:     PaeSheep.cpp
 * Author:   MJW
 * Date:     23.03.2006
 * Modified: 23.03.2006
 *
 * Purpose: Sourcecode zur Klasse PaeSheep
 */
 
#include <math.h>

#include "PaeSheep.h"
#include "Utilities.h"
#include "Definitions.h"
#include "Info.h"


/*
 * Methodendefinitionen zur Klasse PaeSheep
 */



PaeSheep::PaeSheep(void){

  this->SetStandardParameters();
  
}

PaeSheep::PaeSheep(float x, float y, World *the_world){

  this->SetStandardParameters();

  my_world = the_world;
  
  pos[0] = x;
  pos[1] = y;
  pos[2] = my_world->terr->GetHeight(x,y);;

}



PaeSheep::~PaeSheep(){
  gluDeleteQuadric(leg_quadric);  
}



void PaeSheep::Draw(int visual){

  /*
   * festlegen von Standardfarben
   */

  float angle;

  GLfloat white[3] = {1.0,1.0,1.0};
  GLfloat grey[3]  = {0.3,0.3,0.3};


  glMatrixMode(GL_MODELVIEW);
  glPushMatrix();
    glTranslatef(pos[0],pos[1],pos[2]+1.0);
    angle = UtilAngleFromDir(dir);

	glRotatef(angle,0.0,0.0,1.0);
	glScalef(0.5,0.5,0.5);

	/*
	 * Koerper
	 */
    glPushMatrix();
	  if(visual == 0){
        glColor3fv(col);
      } else {
        glColor3fv(vis_col);
	  }
      glScalef(2.0,1.0,1.0);
      glutSolidSphere(1.0,6,6);
    glPopMatrix();


    /*
	 * Schwanz
	 */

    glPushMatrix();    
      glColor3fv(white);
	  glTranslatef(2.1,0.0,0.0);
	  glScalef(0.3,0.3,0.3);
	  glutSolidSphere(1.0,5,5);
    glPopMatrix();
         

    /*
	 * Kopf
	 */

	glPushMatrix();
      glColor3fv(grey);
	  glTranslatef(-1.9,0.0,0.0);
	  if( zustand == GRASEN){
		glTranslatef(0.0,0.0,-0.6);
	  } else {
		glTranslatef(0.0,0.0,0.6);
	  }
	  glScalef(0.7,0.5,0.5);
	  glutSolidSphere(1.0,10,10); 
	glPopMatrix();
         

	glPushMatrix();
      glRotatef(180,1.0,0.0,0.0); 

      // ### front left leg ###
	  glTranslatef(-0.75,0.25,0.9);
	  this->DrawLeg();

      // ### front right leg ###
	  glTranslatef(0.0,-0.5,0.0);
	  this->DrawLeg();

      // ### back left leg ###
	  glTranslatef(1.3,0.5,0.0);
	  this->DrawLeg();

      // ### back right leg ###
	  glTranslatef(0.0,-0.5,0.0);
	  this->DrawLeg();
	glPopMatrix();
             
  glPopMatrix();
}


void PaeSheep::Think(void){
  BasicInfo *b_info;  
  float p[3];
  float dist;
  float gras;
  float sheep_dist;
  float p_mid[3] = {0.0,0.0,0.0};
  float stone_dist;
  int i, num_of_nachbarn=0;
  PaeSheep **nachbarn=NULL;
  PaeSheep *near_sheep;
  Stone * stone_near;
  BasicInfo *stone_info;
  
  float stone_delta[3];
  
  
  /*
   * Wenn das Schaf noch nicht lange
   * sucht, laeuft es erstmal weiter!
   */
  if(zustand == SUCHEN && suchzeit > 0){                 
    suchzeit--;
  } else {
    /*
     * Wieviel Gras gibt es hier??
     */
    gras = my_world->GetGras(pos[0],pos[1]);
  
    /*
     * Grasen oder Suchen?
     */  
    if(gras > GRAS_LEVEL){
      zustand = GRASEN;
      this->SetVisColor(0.0,1.0,0.0);  // Grasendes Schaf ist hellgruen
      return;
    } else {
      if(zustand != SUCHEN){
        suchzeit = SUCHZEIT+(int)PaeRand(-SUCHDELTA,SUCHDELTA);
      }
      zustand = SUCHEN;
    }
  }

  this->SetVisColor(0.0,0.5,0.1);  // Suchendes Schaf ist dunkelgruen
  
  /*
   * Start with new Acc
   */
    
  acc[0] = PaeRand(-0.5,0.5);
  acc[1] = PaeRand(-0.5,0.5);
  acc[2] = 0.0;
  
  nachbarn = my_world->GetFiveNearestSheep(this);
  dist = 0;
  for(i=0;i<5;i++){
    if(nachbarn[i] != NULL){
      sheep_dist = Distance(nachbarn[i],this);
      dist += sheep_dist;
      nachbarn[i]->GetPosition(p);
      p_mid[0] += p[0];
      p_mid[1] += p[1];
      p_mid[2] += p[2];
      num_of_nachbarn++;
      
      if(i == 0 && sheep_dist < 2.0){
        acc[0] = (-p[0]+pos[0])/(sheep_dist+0.1);
        acc[1] = (-p[1]+pos[1])/(sheep_dist+0.1);
        acc[2] = 0.0;
      }
        
    }
  }
  dist /= num_of_nachbarn;
  p_mid[0] /= num_of_nachbarn;
  p_mid[1] /= num_of_nachbarn;
  p_mid[2] /= num_of_nachbarn;
  
  if(dist > HERDENDISTANZ){
    this->SetVisColor(0.6,0.1,0.8);  // Einsames Schaf ist violett
          
    dir[0] = p_mid[0]-pos[0];
    dir[1] = p_mid[1]-pos[1];
    dir[2] = p_mid[2]-pos[2];

    acc[0] += p_mid[0]-pos[0];
    acc[1] += p_mid[1]-pos[1];
    acc[2] += p_mid[2]-pos[2];
  }
  
  stone_near = my_world->GetNearestStone(this);
  stone_info = stone_near->SheepInfo(this);
  stone_dist = stone_info->GetDistance();
  stone_info->GetNormal(stone_delta);
  
  if(stone_dist < 6.0){
    acc[0] += stone_delta[0]/(stone_dist+0.1);
    acc[1] += stone_delta[1]/(stone_dist+0.1);
    acc[2] += 0.0;
  }
  
  
  
  delete(nachbarn);
}
  
void PaeSheep::Move(void){
  
  if (zustand == GRASEN){
    my_world->EatGras(pos[0],pos[1]);
    return;
  }
     
  dir[0] += acc[0];
  dir[1] += acc[1]; 
  dir[2] += acc[2];
    
  (void)PaeRenorm3(dir); 
  
  pos[0] += dir[0]*v;
  pos[1] += dir[1]*v;
  pos[2] += dir[2]*v;
  
}
  

void PaeSheep::SetStandardParameters(void){
  float angle;

  my_world = NULL;

  pos[0] = 0.0;
  pos[1] = 0.0;
  pos[2] = 0.0;
  

  angle = PaeRand(-3.1415926,3.1415926);
  dir[0] = sin(angle);
  dir[1] = cos(angle);
  dir[2] = 0.0;

  acc[0] = 0.0;
  acc[1] = 0.0;
  acc[2] = 0.0;
  
  up[0] = 0.0;
  up[1] = 0.0;
  up[2] = 1.0;
  
  col[0] = 1.0-PaeRand(0.0,0.3);
  col[1] = col[0];
  col[2] = col[0];
  
  vis_col[0] = 0.7;
  vis_col[1] = 0.7;
  vis_col[2] = 0.7;
  
  v = PaeRand(0.2,0.4);

  leg_quadric = gluNewQuadric();
  
  zustand = GRASEN;
  
  nextsheep = NULL;
}

  
void PaeSheep::DrawLeg(void){
  GLfloat white[3] = {1.0,1.0,1.0};
  GLfloat black[3] = {0.1,0.1,0.1};

  gluQuadricDrawStyle(leg_quadric, GLU_FILL);

  glPushMatrix();
    // ### Draw the leg ... ###
    glColor3fv(white);
    gluCylinder(leg_quadric, 0.1, 0.1, 0.5, 4, 1);

     /* ... including hoof ... */
    glTranslatef(0.0,0.0,0.4);
    glColor3fv(black);
    gluCylinder(leg_quadric, 0.11, 0.11, 0.15, 4, 1);
  glPopMatrix();  
}
