/* 
 * File:     PaeSimulator.c
 * Author:   MJW
 * Date:     23.03.2006
 * Modified: 23.03.2006
 *
 * Purpose:  Sourcecode fuer den Simulator
 */

#include "PaeSimulator.h"
#include "wrapper.h"
#include <string.h>

using namespace std;



PaeSimulator::PaeSimulator(void){

  polygonMode = GL_FILL;

  light_position[0] = 49.0;
  light_position[1] = 49.0;
  light_position[2] = 49.0;
  light_position[3] =  0.0;

  light_ambient[0] = 0.2;
  light_ambient[1] = 0.2;
  light_ambient[2] = 0.2;
  light_ambient[3] = 1.0;

  light_diffuse[0] = 1.0;
  light_diffuse[1] = 1.0;
  light_diffuse[2] = 0.7;
  light_diffuse[3] = 1.0;

  light_specular[0] = 1.0;
  light_specular[1] = 1.0;
  light_specular[2] = 1.0;
  light_specular[3] = 1.0;

  win_w = 1000;
  win_h = 700;

  zoom = 1.0;
  
  time_step = 100;

  flat = 0;

  cam = new PaeCamera();
  cam->SetPosition(-20.0,40.0,40.0);
  cam->SetDirection(0.7,0.0,-0.6);

  wor = new World();

  visual_flag = 0;
  mode = SIMULATOR_ACTION;
}


void PaeSimulator::ToggleVisualFlag(void) {
  visual_flag = 1-visual_flag;
}

void PaeSimulator::Display(void) {

  char render_string[5];


    /* FrameBuffer und Z-Buffer loeschen */
    glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);


	glMatrixMode(GL_MODELVIEW);
    glPushMatrix();

  	  glLightfv(GL_LIGHT0, GL_POSITION, light_position);
      /* ************************************** *\
       * /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ *
       *   Hier die Zeichenbefehle einfuegen!!  *
      \* ************************************** */
      wor->Draw(visual_flag);

	glMatrixMode(GL_MODELVIEW);
    glPopMatrix();

	glMatrixMode(GL_MODELVIEW);
    glPushMatrix();
      /*
       * FROM here on it's for the screen text!!!!!
       */
	glDisable(GL_LIGHTING);
      glMatrixMode(GL_PROJECTION);
      glPushMatrix();
        glLoadIdentity();
        gluOrtho2D(0.0,win_w,0.0,win_h);
      
        glMatrixMode(GL_MODELVIEW);
        glLoadIdentity();
      
   
        glColor3f(1.0,1.0,1.0);

        glRasterPos2i(100,100);
        UtilItoa(time_step,render_string,10);
		UtilRenderString("Zeitschrittweite: ");
		UtilRenderString(render_string);

        glRasterPos2i(100,75);
        UtilItoa(wor->GetNumberOfSheep(),render_string,10);
		UtilRenderString("Anzahl Schafe : ");
		UtilRenderString(render_string);

        glRasterPos2i(100,50);
        UtilItoa(wor->GetNumberOfCadavre(),render_string,10);
		UtilRenderString("Anzahl Kadaver: ");
		UtilRenderString(render_string);
		
 
      glMatrixMode(GL_PROJECTION);
      glPopMatrix();

	glMatrixMode(GL_MODELVIEW);
    glPopMatrix();
	glEnable(GL_LIGHTING);



    /*
     * END of the screen text!!!!!
     */


    glutSwapBuffers();
}




void PaeSimulator::Reshape(int width, int height) {

  win_w = width;
  win_h = height;
  
  /* Projection Transformation */
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluPerspective(60.0,(float)width/(float)height,1.0,200.0);
  glViewport(0, 0, (GLint)width, (GLint)height);
	
  /* Viewing Transformation */
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity ();
  cam->SetCam();

  glutPostRedisplay();
}


void PaeSimulator::Init(int argc, char **argv) {

  /*
   * Setting up the window
   */

  glutInit(&argc, argv);
  glutInitWindowPosition(0, 0);
  glutInitWindowSize(win_w,win_h);
  glutInitDisplayMode(GLUT_DEPTH | GLUT_RGBA | GLUT_DOUBLE);
  
  if(glutCreateWindow("Digitale Schafweide") == GL_FALSE){
	exit(1);
  }

  /* alpha-testing */
  glBlendFunc(GL_SRC_ALPHA, GL_ONE);  
     
  /* z-Buffering */
  glEnable(GL_DEPTH_TEST);

  /* Color => Material */
  glEnable(GL_COLOR_MATERIAL); 

  /* Normals are normalized automatically */
  glEnable(GL_NORMALIZE);

  /* Set light properties */ 
  glLightfv(GL_LIGHT0, GL_AMBIENT, light_ambient);
  glLightfv(GL_LIGHT0, GL_DIFFUSE, light_diffuse);
  glLightfv(GL_LIGHT0, GL_SPECULAR, light_specular);		

  /* Enable lighting computation and light0 */  
  glEnable(GL_LIGHTING);
  glEnable(GL_LIGHT0);

  Reshape(win_w,win_h);

}





void PaeSimulator::Key(unsigned char key, int x, int y) {

  /* Switch lighting calculation on/off */
  switch(key) {

  case 'w':
    wor->terr->SwitchWater();
    break;

  case 'm':
    wor->terr->AddHeight(0.1);
    break;
  case 'M':
    wor->terr->AddHeight(-0.1);
    break;

  case 't':
    time_step -= 4;
    UtilClamp(&time_step,10,1000);
    glutPostRedisplay();
    break;
  case 'T':
    time_step += 4;
    UtilClamp(&time_step,10,1000);
    glutPostRedisplay();
    break;

  case 'h':
    cam->RotateDir(0.1);
    cam->SetCam();
    break;
  case 'H':
    cam->RotateDir(-0.1);
    cam->SetCam();
    break;
    
  case 'g':
    cam->Rotate(0.1);
    cam->SetCam();
    break;
  case 'G':
    cam->Rotate(-0.1);
    cam->SetCam();
    break;

  case 'p':
    cam->Pitch(0.1);
    cam->SetCam();
    break;
  case 'P':
    cam->Pitch(-0.1);
    cam->SetCam();
    break;

  case 'y':
    cam->Yaw(0.1);
    cam->SetCam();
    break;
  case 'Y':
    cam->Yaw(-0.1);
    cam->SetCam();
    break;

  case 'r':
    cam->Roll(0.1);
    cam->SetCam();
    break;
  case 'R':
    cam->Roll(-0.1);
    cam->SetCam();
    break;

  case 's':
    wor->terr->SoftenTerrain();
    wor->terr->ReBuildNormals();
    break;

  case 'o':
    cam->StatusPrint();
    break;

  case 'l':
	if(glIsEnabled(GL_LIGHTING)) {
	  glDisable(GL_LIGHTING);
	} else {
	  glEnable(GL_LIGHTING);
	}
	glutPostRedisplay ();
    break;

  case 'a':
    if (mode == SIMULATOR_ACTION){
	  mode = SIMULATOR_SLEEP;
	} else {
	  mode = SIMULATOR_ACTION;
	}
	glutPostRedisplay ();
    break;

  case 27:
  case 'q':
	exit(1);

  case 'v':
    this->ToggleVisualFlag();
	glutPostRedisplay ();
    break;

  case 'f':
    flat = 1-flat;
	cout << endl << flat;
	if (flat > 0){
	  glShadeModel(GL_FLAT);
	} else {
	  glShadeModel(GL_SMOOTH);
    }
	glutPostRedisplay ();
	break;

  case ' ':
	if(polygonMode==GL_FILL){
	  polygonMode=GL_LINE;
    } else {
	  polygonMode=GL_FILL;
    }
    glPolygonMode (GL_FRONT_AND_BACK, polygonMode);

	glutPostRedisplay ();
	break;

  default:
    break;
  }
}


void PaeSimulator::SpecialKey(int key, int x, int y) {

  /* Model rotation via arrow keys */
  switch(key) {
  case GLUT_KEY_UP:
    cam->MoveForward(1.0);
	break;
  case GLUT_KEY_DOWN:
    cam->MoveForward(-1.0);
	break;
  case GLUT_KEY_LEFT:
    cam->MoveSide(1.0);
	break;
  case GLUT_KEY_RIGHT:
    cam->MoveSide(-1.0);
	break;
  case GLUT_KEY_PAGE_UP:
    cam->MoveUp(1.0);
    break;
  case GLUT_KEY_PAGE_DOWN:
    cam->MoveUp(-1.0);
    break;  
  default:
    break;
  }

  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  cam->SetCam();
  glutPostRedisplay();
}


void PaeSimulator::Timer(int value){


  if(mode == SIMULATOR_ACTION){
	wor->Prepare();
	
	wor->Action();
  }

  glutPostRedisplay();
  glutTimerFunc(time_step, WrapTimer, value+1);
}

