/* 
 * File:     Terrain.cpp
 * Author:   PAE
 * Date:     18.03.2006
 * Modified: 23.03.2006
 *
 * Purpose: Sourcecode zur Klasse Terrain
 */
 
#include "Terrain.h"
#include "Definitions.h"

//routinen zur terrain klasse

void Terrain::DrawWater(void){
glDepthMask(GL_FALSE);
glEnable(GL_BLEND);
  glBegin(GL_QUADS);
    glNormal3f(0.0,0.0,1.0);
    glColor4f(0.0,0.0,1.0,0.6);
    glVertex3f(0.0,0.0,WASSERHOEHE);
    glVertex3f(0.0,height-1,WASSERHOEHE);
    glVertex3f(width-1,height-1,WASSERHOEHE);
    glVertex3f(width-1,0.0,WASSERHOEHE);
  glEnd();
glDisable(GL_BLEND);
glDepthMask(GL_TRUE);
}



Terrain::Terrain(){
  int i, j, n;  //zaehlvariablen
  float c_x,c_y;
  float m;   //fuer die gerade durch das terrain f(x) = m*x+c
  float *f_x;   //hier werden die funktionswerte gespeichert
  float fx;
  float delta;  // um soviel wird die eine Haelfte erhoeht und die andere erniedrigt

  float max_diff,top,bot;

  water  = false;
  width  = 200;
  height = 200;

  max_diff = HOEHENUNTERSCHIED;

  scal_x = 1.0;
  scal_y = 1.0;

  shift_x = 50.0;
  shift_y = 50.0;

  //speicherplatz wird reserviert fr die terrain daten 
  data = new float**[width];
  for(i = 0; i < width; i++){
    data[i] = new float*[height];
    for(j = 0; j < height; j++){
      data[i][j] = new float[2];
      data[i][j][0] = STARTHOEHE;//hoehe = STARTHOEHE
      data[i][j][1] = PaeRand(0.0,1.0);   
    }
  } 
  cout << "Speicherplatz fuer das Terrain bereitgestellt." << endl;
  //die hoehe wird jetzt erzeugt mithilfe des Zerteilungsalgorithmus
  //d.h., dass eine gerade durch das terrain gezogen wird, 
  //und alle hoehenwerte auf der einen seite um einen kleinen wert 
  //erhoeht werden, alle auf der anderen seite um denselben kleinen wert
  //erniedrigt werden
  
  f_x = new float[width];
  
  for(n = 0; n < SCHRITTZAHL; n++){
    m = PaeRand(-89.99, 89.99);//der steigungswinkel der geraden
    m = tan(m / 180.0 * M_PI);//umgerechnet in eine zahl
    c_x = PaeRand(0.0,1.0)*(width-1.0);
    c_y = PaeRand(0.0,1.0)*(height-1.0);
	delta = PaeRand(-1.0,1.0)*PaeRand(-1.0,1.0)*PaeRand(-1.0,1.0)*max_diff;

    for(i = 0; i < width; i++){
	  fx = c_y-(c_x-i)*m;
      for(j = 0; j < height; j++){
        if(j > fx){
          data[i][j][0] += delta;
        }else{
          data[i][j][0] -= delta;
        }
      }
    }
  }  
  cout << "Terrain erzeugt." << endl;



  // hoechsten und tiefsten Terrainwert suchen  
  top = data[0][0][0];
  bot = data[0][0][0];
  
  for(i = 0; i < width; i++){
    for(j = 0; j < height; j++){
      if(data[i][j][0] > top){
        top = data[i][j][0];
      }
      if(data[i][j][0] < bot){
        bot = data[i][j][0];
      }
    }
  }

  // Terrain skalieren auf [-RANGE,RANGE]
  if (top-bot > 0.0){
    for(i = 0; i < width; i++){
      for(j = 0; j < height; j++){
        data[i][j][0]  = 2.0*(data[i][j][0]-bot)/(top-bot) - 1.0;
        data[i][j][0] *= RANGE;
      }
    }
  }
               

    //hier wird der speicherplatz fuer die normalen bereitgestellt
    normals = new float**[width];
    for(i = 0; i < width; i++){
      normals[i] = new float*[height];
      for(j = 0; j < height; j++){
        normals[i][j] = new float[3];
        normals[i][j][0] = 0.0;
        normals[i][j][1] = 0.0;
        normals[i][j][2] = 1.0;
      }
    }
    cout << "Speicherplatz fuer die Normalen bereitgestellt." << endl;

	this->ReBuildNormals();

}


float Terrain::GetFood(float wx, float wy){
  int links = 0, rechts = 0, oben = 0, unten = 0;
  float d_x = 0.0, d_y = 0.0, food_left = 0.0, food_right = 0.0;

  float x,y;
  TransformWorldTerrain(wx,wy,&x,&y);
  
  //liegt die stelle noch im terrain ? der rand zaehlt als nicht mehr drin
  if(width-1 < x || height-1 < y || x < 0 || y < 0){
    return 0.0;
  }
  
  //die koordinaten um x|y aussenrum
  links = (int)x;
  rechts = (int)(x+1.0);
  oben = (int)(y+1.0);
  unten = (int)y;
  
  d_x = rechts - x;
  d_y = oben - y;
  
  food_left = d_y * data[links][unten][1] + (1 - d_y) * data[links][oben][1];
  food_right = d_y * data[rechts][unten][1] + (1 - d_y) * data[rechts][oben][1];
  
  return d_x * food_left + (1 - d_x) * food_right;
}


void Terrain::ReduceFood(float wx, float wy, float amount){
  float food;
  float percentage = 0.5;
  int links = 0, rechts = 0, oben = 0, unten = 0;
  
  float x,y;
  TransformWorldTerrain(wx,wy,&x,&y);
  food = GetFood(x, y);

  //wenn food <= 0 dann liegt die stelle entweder nicht im terrain oder es kann
  //sowieso nicht reduziert werden
  if(food > 0.0){
  
    //die koordinaten um x|y aussenrum
    links = (int)x;
    rechts = (int)(x+1.0);
    oben = (int)(y+1.0);
    unten = (int)y;
    
    //wieviel prozent werden gefressen ?
    percentage = amount / food;
    //bei jedem der punkte aussenrum wird der pflanzenanteil reduziert um percentage
    data[links][unten][1]  = max((1.0-percentage)*data[links][unten][1],0.0);
    data[rechts][unten][1] = max((1.0-percentage)*data[rechts][unten][1],0.0);
    data[links][oben][1]   = max((1.0-percentage)*data[links][oben][1],0.0);
    data[rechts][oben][1]  = max((1.0-percentage)*data[rechts][oben][1],0.0);
  }    
}


float Terrain::GetHeight(float wx, float wy){
  int links, rechts, oben, unten;
  float d_x, d_y, z_left, z_right;

  float x,y;
  TransformWorldTerrain(wx,wy,&x,&y);

  //liegt die stelle noch im terrain ? der rand zaehlt als nicht mehr drin
  if(width-1 < x || height-1 < y || x < 0 || y < 0){
    return 0.0;
  }
  //die koordinaten der punkte um die stelle x|y aussenrum
  links = (int)x;
  rechts = (int)(x+1.0);
  oben = (int)(y+1.0);
  unten = (int)y;
  
  d_x = rechts - x;
  d_y = oben - y;
  
  z_left = d_y * data[links][unten][0] + (1 - d_y) * data[links][oben][0];
  z_right = d_y * data[rechts][unten][0] + (1 - d_y) * data[rechts][oben][0];

  return d_x * z_left + (1 - d_x) * z_right;
}



void Terrain::GrowFood(void){
  int i, j;
  float food = 0.0;  
  const float k = GRASWACHSTUMSKONSTANTE;

  for(i = 0; i < width; i++){
    for(j = 0; j < height; j++){
      food = data[i][j][1];
      //wenn genug essen da ist, wchst es logistisch
      if(food >= 0.001){
        data[i][j][1] = (food + k * food * (1 - food));
      //sonst wird es auf genug gesetzt
      }else{
        data[i][j][1] = 0.002;
      }
    }
  }
}


void Terrain::Draw(void){
  int i, j;
  float del0,del1;  
  
  int col[2][3] = { {158,120, 18},
					{ 86,207, 56}};

  glPushMatrix();
    glTranslatef(-shift_x,-shift_y,0.0);
    glScalef(scal_x,scal_y,1.0);
    for(i = 0; i < width-1; i++){
      glBegin(GL_TRIANGLE_STRIP);
      for(j = 0; j < height; j++){
		del0 = data[i  ][j][1];
		del1 = data[i+1][j][1];
        glNormal3fv(normals[i][j]);
        glColor3ub((GLubyte)((1-del0)*col[0][0] + del0*col[1][0]),
				   (GLubyte)((1-del0)*col[0][1] + del0*col[1][1]),
				   (GLubyte)((1-del0)*col[0][2] + del0*col[1][2]));
        glVertex3f(i,j,data[i][j][0]);

        glNormal3fv(normals[i+1][j]);
        glColor3ub((GLubyte)((1-del1)*col[0][0] + del1*col[1][0]),
				   (GLubyte)((1-del1)*col[0][1] + del1*col[1][1]),
				   (GLubyte)((1-del1)*col[0][2] + del1*col[1][2]));
        glVertex3f(i+1,j,data[i+1][j][0]);
      }
      glEnd();
    }
    if(water){
      DrawWater();
    }
  glPopMatrix();
}


//das wasser wird an/ausgeschaltet
void Terrain::SwitchWater(void){
  water = !water;
}


void Terrain::AddHeight(float amount){
int i, j;
  //jeder punkt im terrain wird um amount erhoeht
  for(i = 0; i < width; i++){
    for(j = 0; j < height; j++){
      data[i][j][0] += amount;
    }
  }
}



void Terrain::ReBuildNormals(void){
  int i, j;
  float v1[3], v2[3], nv[3];

  // Randnormalen setzen
  for(j = 0; j < height; j++){
	normals[0][j][0] = 0.0;
	normals[0][j][1] = 0.0;
	normals[0][j][2] = 1.0;
	normals[width-1][j][0] = 0.0;
	normals[width-1][j][1] = 0.0;
	normals[width-1][j][2] = 1.0;
  }

  for(i = 0; i < width; i++){
	normals[i][0][0] = 0.0;
	normals[i][0][1] = 0.0;
	normals[i][0][2] = 1.0;
	normals[i][height-1][0] = 0.0;
	normals[i][height-1][1] = 0.0;
	normals[i][height-1][2] = 1.0;
  }

  //jetzt werden die inneren normalen berechnet und gespeichert
  for(i = 1; i < width-1; i++){
	for(j = 1; j < height-1; j++){
	  //das dreieck links unterhalb des punktes
	  v1[0] = 1.0;
	  v1[1] = 0.0;
	  v1[2] = data[i][j][0] - data[i-1][j][0];
	  
	  v2[0] = 0.0;
	  v2[1] = 1.0;
	  v2[2] = data[i][j][0] - data[i][j-1][0];
        
	  nv[0] = v1[1] * v2[2] - v2[1] * v1[2];
	  nv[1] = v1[2] * v2[0] - v2[2] * v1[0];
	  nv[2] = v1[0] * v2[1] - v2[0] * v1[1];
        
	  normals[i][j][0] = nv[0]/4;
	  normals[i][j][1] = nv[1]/4;
	  normals[i][j][2] = nv[2]/4;
      
	  //das dreieck rechts unterhalb des punktes
	  v1[0] = 0.0;
	  v1[1] = 1.0;
	  v1[2] = data[i][j][0] - data[i-1][j][0];
        
	  v2[0] = -1.0;
	  v2[1] = 1.0;
	  v2[2] = data[i][j][0] - data[i+1][j-1][0];
        
	  nv[0] = v1[1] * v2[2] - v2[1] * v1[2];
	  nv[1] = v1[2] * v2[0] - v2[2] * v1[0];
	  nv[2] = v1[0] * v2[1] - v2[0] * v1[1];
        
	  normals[i][j][0] += nv[0]/8;
	  normals[i][j][1] += nv[1]/8;
	  normals[i][j][2] += nv[2]/8;

	  //das dreieck rechts des punktes
	  v1[0] = -1.0;
	  v1[1] = 1.0;
	  v1[2] = data[i][j][0] - data[i+1][j-1][0];
	  
	  v2[0] = -1.0;
	  v2[1] = 0.0;
	  v2[2] = data[i][j][0] - data[i-1][j][0];
	  
	  nv[0] = v1[1] * v2[2] - v2[1] * v1[2];
	  nv[1] = v1[2] * v2[0] - v2[2] * v1[0];
	  nv[2] = v1[0] * v2[1] - v2[0] * v1[1];
        
	  normals[i][j][0] += nv[0]/8;
	  normals[i][j][1] += nv[1]/8;
	  normals[i][j][2] += nv[2]/8;

	  //das dreieck rechts oberhalb des punktes
	  v1[0] = -1.0;
	  v1[1] = 0.0;
	  v1[2] = data[i][j][0] - data[i-1][j][0];
        
	  v2[0] = 0.0;
	  v2[1] = -1.0;
	  v2[2] = data[i][j][0] - data[i][j+1][0];
        
	  nv[0] = v1[1] * v2[2] - v2[1] * v1[2];
	  nv[1] = v1[2] * v2[0] - v2[2] * v1[0];
	  nv[2] = v1[0] * v2[1] - v2[0] * v1[1];
        
	  normals[i][j][0] += nv[0]/4;
	  normals[i][j][1] += nv[1]/4;
	  normals[i][j][2] += nv[2]/4;

	  //das dreieck links oberhalb des punktes
	  v1[0] = 0.0;
	  v1[1] = -1.0;
	  v1[2] = data[i][j][0] - data[i][j+1][0];
        
	  v2[0] = 1.0;
	  v2[1] = -1.0;
	  v2[2] = data[i][j][0] - data[i-1][j+1][0];
        
	  nv[0] = v1[1] * v2[2] - v2[1] * v1[2];
	  nv[1] = v1[2] * v2[0] - v2[2] * v1[0];
	  nv[2] = v1[0] * v2[1] - v2[0] * v1[1];
        
	  normals[i][j][0] += nv[0]/8;
	  normals[i][j][1] += nv[1]/8;
	  normals[i][j][2] += nv[2]/8;

	  //das dreieck links des punktes
	  v1[0] = 1.0;
	  v1[1] = -1.0;
	  v1[2] = data[i][j][0] - data[i-1][j+1][0];
        
	  v2[0] = 1.0;
	  v2[1] = 0.0;
	  v2[2] = data[i][j][0] - data[i-1][j][0];
	  
	  nv[0] = v1[1] * v2[2] - v2[1] * v1[2];
	  nv[1] = v1[2] * v2[0] - v2[2] * v1[0];
	  nv[2] = v1[0] * v2[1] - v2[0] * v1[1];
        
	  normals[i][j][0] += nv[0]/8;
	  normals[i][j][1] += nv[1]/8;
	  normals[i][j][2] += nv[2]/8;
      }
    }
   
    cout << "Normalen berechnet." << endl;
}


int Terrain::GetWidth(void){
  return width;
}


int Terrain::GetHeight(void){
  return height;
}


void Terrain::ScaleTerrain(float min, float max){
int i, j;
float t_min, t_max, t_diff, diff;

  diff = max - min;
  t_min = data[0][0][0];
  t_max = data[0][0][0];
  
  for(i = 0; i < width; i++){
    for(j = 0; j < height; j++){
      if(t_min > data[i][j][0]){
        t_min = data[i][j][0];
      }
      if(t_max < data[i][j][0]){
        t_max = data[i][j][0];
      }
    }
  }
  
  t_diff = t_max - t_min;
  
  for(i = 0; i < width; i++){
    for(j = 0; j < height; j++){
      data[i][j][0] = (data[i][j][0] - t_min) / t_diff * diff + min;
    }
  }  
    
}


void Terrain::SoftenTerrain(void){
float **temp;//in diesem array werden die neuen hoehendaten zwischengespeichert
int i, j;

  temp = new float*[width];
  for(i = 0; i < width; i++){
    temp[i] = new float[height];
    for(j = 0; j < height; j++){
      temp[i][j] = 0.0;    
    }
  }  
  
  for(i = 1; i < width-1; i++){//die randpunkte werden extra berechnet
    for(j = 1; j < height-1; j++){
      temp[i][j] = 0.6 * data[i][j][0];//60% der neuen hoehe ist die alte hoehe
      //die restlichen 40% verteilen sich auf die umliegenden 8 punkte, 
      //daher werden die randpunkte extra berechnet
      //links oben
      temp[i][j] += 0.05 * data[i-1][j+1][0];
      //oben
      temp[i][j] += 0.05 * data[i][j+1][0];
      //rechts oben
      temp[i][j] += 0.05 * data[i+1][j+1][0];
      //rechts
      temp[i][j] += 0.05 * data[i+1][j][0];
      //rechts unten
      temp[i][j] += 0.05 * data[i+1][j-1][0];
      //unten
      temp[i][j] += 0.05 * data[i][j-1][0];
      //links unten
      temp[i][j] += 0.05 * data[i-1][j-1][0];
      //links
      temp[i][j] += 0.05 * data[i-1][j][0];
    }
  }
  
  //die unteren und oberen randpunkte werden berechnet
  for(i = 1; i < width-1; i++){
    temp[i][0] = 0.75 * data[i][0][0];//75% der neuen hoehe ist die alten hoehe
    //links
    temp[i][0] += 0.05 * data[i-1][0][0];
    //links oben
    temp[i][0] += 0.05 * data[i-1][1][0];
    //oben
    temp[i][0] += 0.05 * data[i][1][0];
    //rechts oben
    temp[i][0] += 0.05 * data[i+1][1][0];
    //rechts
    temp[i][0] += 0.05 * data[i+1][0][0]; 
    
    temp[i][height-1] = 0.75 * data[i][height-1][0];//75% der neuen hoehe ist die alten hoehe
    //links
    temp[i][height-1] += 0.05 * data[i-1][height-1][0];
    //links unten
    temp[i][height-1] += 0.05 * data[i-1][height-2][0];
    //unten
    temp[i][height-1] += 0.05 * data[i][height-2][0];
    //rechts unten
    temp[i][height-1] += 0.05 * data[i+1][height-2][0];
    //rechts
    temp[i][height-1] += 0.05 * data[i+1][height-1][0]; 
  }
  
  //die linken und rechten randpunkte werden berechnet
  for(j = 1; j < width-1; j++){
    temp[0][j] = 0.75 * data[0][j][0];//75% der neuen hoehe ist die alte hoehe
    //unten
    temp[0][j] += 0.05 * data[0][j-1][0];
    //rechts unten
    temp[0][j] += 0.05 * data[1][j-1][0];
    //rechts
    temp[0][j] += 0.05 * data[1][j][0];
    //rechts oben
    temp[0][j] += 0.05 * data[1][j+1][0];
    //oben
    temp[0][j] += 0.05 * data[0][j+1][0];
    
    temp[width-1][j] = 0.75 * data[width-1][j][0];//75% der neuen hoehe ist die alte hoehe
    //unten
    temp[width-1][j] += 0.05 * data[width-1][j-1][0];
    //links unten
    temp[width-1][j] += 0.05 * data[width-2][j-1][0];
    //links
    temp[width-1][j] += 0.05 * data[width-2][j][0];
    //links oben
    temp[width-1][j] += 0.05 * data[width-2][j+1][0];
    //oben
    temp[width-1][j] += 0.05 * data[width-1][j+1][0];
  }
  
  //die eckpunkte
  
  temp[0][0] = 0.85 * data[0][0][0];
  //rechts
  temp[0][0] += 0.05 * data[1][0][0];
  //rechts oben
  temp[0][0] += 0.05 * data[1][1][0];
  //oben
  temp[0][0] += 0.05 * data[0][1][0];
  
  temp[width-1][0] = 0.85 * data[width-1][0][0];
  //links
  temp[width-1][0] += 0.05 * data[width-2][0][0];
  //links oben
  temp[width-1][0] += 0.05 * data[width-2][1][0];
  //oben
  temp[width-1][0] += 0.05 * data[width-1][1][0];
  
  temp[width-1][height-1] = 0.85 * data[width-1][height-1][0];
  //links
  temp[width-1][height-1] += 0.05 * data[width-2][height-1][0];
  //links unten
  temp[width-1][height-1] += 0.05 * data[width-2][height-2][0];
  //unten
  temp[width-1][height-1] += 0.05 * data[width-1][height-2][0];
  
  temp[0][height-1] = 0.85 * data[0][height-1][0];
  //rechts
  temp[0][height-1] += 0.05 * data[1][height-1][0];
  //rechts unten
  temp[0][height-1] += 0.05 * data[1][height-2][0];
  //unten
  temp[0][height-1] += 0.05 * data[0][height-2][0];
  
    
  //die neuen werte werden eingetragen
  for(i = 0; i < width; i++){
    for(j = 0; j < height; j++){
      data[i][j][0] = temp[i][j];
    }
  }
  
  //der speicherplatz wird wieder freigegeben
  for(i = 0; i < width; i++){
    delete temp[i];
  }
  delete temp;

  cout << "Terrain weicher gemacht" << endl;
}


//routinen zur uhr

PaeClock::PaeClock(void){
  minutes = 0;
  hours = 0;
  days = 1;
  months = 1;
  years = 0;
}


void PaeClock::Tick(void){
  minutes++;
  if(minutes % 60 == 0){
    minutes = 0;
    hours++;
    if(hours % 24 == 0){
      hours = 0;
      days++;
      if(days % 31 == 0){
        days = 1;
        months++;
        if(months % 13 == 0){
          months = 1;
          years++;
        }
      }
    }
  }
}


int PaeClock::WhichTime(int what){
  switch(what){
    case MINUTES:
      return minutes;
      break;
    case HOURS:
      return hours;
      break;
    case DAYS:
      return days;
      break;
    case MONTHS:
      return months;
      break;
    case YEARS:
      return years;
      break;
  }
  return(0);
}



void PaeClock::PrintTime(void){
  printf("%2i:%2i\t%2i.%2i.%2i\n", hours, minutes, days, months, years);
}



void PaeClock::Set(int what, int amount){
  switch(what){
    case MINUTES:
      minutes = amount % 60;
      break;
    case HOURS:
      hours = amount % 24;
      break;
    case DAYS:
      days = amount % 31 + 1;
      break;
    case MONTHS:
      months = amount % 13 + 1;
      break;
    case YEARS:
      years = amount;
      break;
  }
}


void Terrain::TransformWorldTerrain(float wx, float wy, float *tx, float *ty){
  *tx = (wx+shift_x)/scal_x;
  *ty = (wy+shift_y)/scal_y;
}

void Terrain::TransformTerrainWorld(float tx, float ty, float *wx, float *wy){
  *wx = tx*scal_x-shift_x;
  *wy = ty*scal_y-shift_y;
}
